package ca.recrown.islandsurvivalcraft.world.Information;

import java.util.HashSet;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ThreadFactory;

import ca.recrown.islandsurvivalcraft.utilities.GeneralUtilities;
import ca.recrown.islandsurvivalcraft.utilities.datatypes.Point2;
import ca.recrown.islandsurvivalcraft.world.IslandWorldMap;

public class IslandInformationManager {
    private final IslandWorldMap islandMap;
    private final ExecutorService executors = Executors.newFixedThreadPool(2, new ThreadFactory() {
        @Override
        public Thread newThread(Runnable r) {
            Thread thread = new Thread(r);
            thread.setName("ISC-Chunk-Info-Loader");
            return thread;
        }
    });
    private final ConcurrentHashMap<Point2, Future<Void>> building = new ConcurrentHashMap<>();
    private final ConcurrentHashMap<Point2, HashSet<IslandInformation>> islandSets = new ConcurrentHashMap<>();

    public IslandInformationManager(IslandWorldMap islandWorldMap) {
        this.islandMap = islandWorldMap;
    }

    public void loadChunkIslandInformation(Point2 chunkCoords) {
        Point2[] surrounding = new Point2[4];
        surrounding[0] = chunkCoords.shift(1, 0);
        surrounding[1] = chunkCoords.shift(-1, 0);
        surrounding[2] = chunkCoords.shift(0, -1);
        surrounding[3] = chunkCoords.shift(0, 1);
        HashSet<Point2> checkedCoordinates = new HashSet<>();
        for (Point2 chunk : surrounding) {
            HashSet<IslandInformation> chunkIslandSet = islandSets.computeIfAbsent(chunk, (p) -> new HashSet<>());
            for (IslandInformation islandInformation : chunkIslandSet) {
                if (islandInformation.isIslandInChunk(chunkCoords)) {
                    checkedCoordinates.addAll(islandInformation.getIslandCoordinates());
                }
            }
        }

        for (int x = 0; x < GeneralUtilities.CHUNK_SIZE; x++) {
            for (int z = 0; z < GeneralUtilities.CHUNK_SIZE; z++) {
                int worldX = chunkCoords.x * GeneralUtilities.CHUNK_SIZE + x;
                int worldZ = chunkCoords.y * GeneralUtilities.CHUNK_SIZE + z;
                if (islandMap.isIsland(worldX, worldZ)) {
                    IslandInformationBuilder infoBuilder = new IslandInformationBuilder();
                    Point2 origin = islandMap.findIslandOrigin(worldX, worldZ, (p) -> {
                        if (!checkedCoordinates.add(p))
                            return true;
                        infoBuilder.addCoordinate(p, islandMap.isEdgeOfIsland(p.x, p.y));
                        infoBuilder.addChunk(GeneralUtilities.worldToChunkCoordinates(p));
                        return false;
                    });
                    if (origin != null) {
                        IslandInformation islandInfo = infoBuilder.build(origin);
                        for (Point2 chunk : infoBuilder.getChunksUsed()) {
                            islandSets.computeIfAbsent(chunk, (p) -> new HashSet<>()).add(islandInfo);
                        }
                    }
                }
            }
        }
    }

    public void loadChunkIslandInformationAsync(Point2 chunkCoords) {
        building.put(chunkCoords, executors.submit(() -> {
            loadChunkIslandInformation(chunkCoords);
            return null;
        }));
    }

    public void unloadChunkIslandInformation(Point2 chunkCoords) {
        islandSets.remove(chunkCoords);
        building.remove(chunkCoords);
    }

    public HashSet<IslandInformation> getChunkIslandInformationSet(Point2 chunkCoords, boolean checkBuilding) {
        if (!islandSets.containsKey(chunkCoords)) {
            if (checkBuilding && building.containsKey(chunkCoords)) {
                try {
                    building.get(chunkCoords).get();
                } catch (InterruptedException | ExecutionException e) {
                }
            }
        }
        building.remove(chunkCoords);
        return islandSets.get(chunkCoords);
    }

    public IslandInformation getIslandInformation(Point2 coords) {
        HashSet<IslandInformation> chunkIslandSet = getChunkIslandInformationSet(GeneralUtilities.worldToChunkCoordinates(coords), true);
        for (IslandInformation islandInformation : chunkIslandSet) {
            if (islandInformation.isWithinIsland(coords)) return islandInformation;
        }
        return null;
    }
}